# -*- coding: utf-8 -*-
#
# Copyright 2017-2021- Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test workflow commands."""

# from renku.core.incubation.command import Command, CommandResult
from renku.cli import cli


def test_workflow_remove_command(runner, project):
    """test workflow remove with builder."""
    # TODO: add fixture project with graph generated by default
    result = runner.invoke(cli, ["graph", "generate"])
    assert 0 == result.exit_code

    workflow_name = "test_workflow"

    result = runner.invoke(cli, ["graph", "workflow", "remove", workflow_name])
    assert 2 == result.exit_code

    result = runner.invoke(cli, ["run", "--success-code", "0", "--no-output", "--name", workflow_name, "echo", "foo"])
    assert 0 == result.exit_code

    result = runner.invoke(cli, ["graph", "workflow", "remove", "--force", workflow_name])
    assert 0 == result.exit_code



# -*- coding: utf-8 -*-
#
# Copyright 2018-2021 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Classes for tracking parameter values in provenance."""

from pathlib import Path
from typing import Any, Union
from uuid import uuid4

from marshmallow import EXCLUDE

from renku.core.models.calamus import JsonLDSchema, Nested, fields, prov, renku, schema
from renku.core.models.workflow.parameter import (
    CommandInput,
    CommandInputSchema,
    CommandOutput,
    CommandOutputSchema,
    CommandParameter,
    CommandParameterSchema,
)


class ParameterValue:
    """Value for a parameter in provenance."""

    def __init__(self, *, id: str):
        self.id = id

    @staticmethod
    def generate_id(activity_id: str) -> str:
        """Generate a default id."""
        return f"{activity_id}/parameter-value/{uuid4()}"


class PathParameterValue(ParameterValue):
    """Value for a path parameter in provenance."""

    def __init__(self, *, id: str, parameter: Union[CommandInput, CommandOutput], path: Union[Path, str]):
        super().__init__(id=id)
        self.parameter: Union[CommandInput, CommandOutput] = parameter
        self.path: Union[Path, str] = str(path)


class VariableParameterValue(ParameterValue):
    """Value for a parameter in provenance."""

    def __init__(self, *, id: str, parameter: CommandParameter, value: Any):
        super().__init__(id=id)
        self.parameter: CommandParameter = parameter
        self.value: Any = value


class ParameterValueSchema(JsonLDSchema):
    """ParameterValue schema."""

    class Meta:
        """Meta class."""

        rdf_type = [renku.ParameterValue, schema.PropertyValue]
        model = ParameterValue
        unknown = EXCLUDE

    id = fields.Id()


class PathParameterValueSchema(ParameterValueSchema):
    """PathParameterValue schema."""

    class Meta:
        """Meta class."""

        rdf_type = [renku.PathParameterValue, schema.PropertyValue]
        model = PathParameterValue
        unknown = EXCLUDE

    parameter = Nested(schema.valueReference, [CommandInputSchema, CommandOutputSchema], reverse=True)
    path = fields.String(prov.atLocation)


class VariableParameterValueSchema(ParameterValueSchema):
    """VariableParameterValue schema."""

    class Meta:
        """Meta class."""

        rdf_type = [renku.VariableParameterValue, schema.PropertyValue]
        model = VariableParameterValue
        unknown = EXCLUDE

    parameter = Nested(schema.valueReference, CommandParameterSchema, reverse=True)
    value = fields.Raw(schema.value)



# -*- coding: utf-8 -*-
#
# Copyright 2018-2021- Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""PoC command for testing the new graph design."""

import click

from renku.cli.utils.callback import ClickCallback
from renku.cli.utils.click import CaseInsensitiveChoice
from renku.core.incubation.command import Command
from renku.core.incubation.graph import (
    FORMATS,
    add_to_dataset,
    create_dataset,
    export_graph,
    generate_graph,
    remove_workflow,
)
from renku.core.incubation.graph import status as get_status
from renku.core.incubation.graph import update as perform_update
from renku.core.models.workflow.dependency_graph import DependencyGraph
from renku.core.utils.contexts import measure


@click.group(hidden=True)
def graph():
    """Proof-of-Concept command for testing the new graph design."""


@graph.command()
@click.option("-f", "--force", is_flag=True, help="Delete existing metadata and regenerate all.")
def generate(force):
    """Create new graph metadata."""

    communicator = ClickCallback()
    generate_graph().with_communicator(communicator).build().execute(force=force)

    click.secho("\nOK", fg="green")


@graph.command()
@click.pass_context
def status(ctx):
    r"""Equivalent of `renku status`."""

    communicator = ClickCallback()
    result = get_status().with_communicator(communicator).build().execute()

    stales, modified, deleted = result.output

    if not modified and not deleted:
        click.secho("Everything is up-to-date.", fg="green")
        return

    if stales:
        click.echo(
            f"Outdated outputs({len(stales)}):\n"
            "  (use `renku log [<file>...]` to see the full lineage)\n"
            "  (use `renku update [<file>...]` to generate the file from its latest inputs)\n"
        )
        for k, v in stales.items():
            paths = click.style(", ".join(sorted(v)), fg="red", bold=True)
            click.echo(f"\t{k}:{paths}")
        click.echo()
    else:
        click.secho("All files were generated from the latest inputs.", fg="green")

    if modified:
        click.echo(
            f"Modified inputs({len(modified)}):\n"
            "  (use `renku log --revision <sha1> <file>` to see a lineage for the given revision)\n"
        )
        for v in modified:
            click.echo(click.style(f"\t{v}", fg="blue", bold=True))
        click.echo()

    if deleted:
        click.echo(
            "Deleted files used to generate outputs:\n"
            "  (use `git show <sha1>:<file>` to see the file content for the given revision)\n"
        )
        for v in deleted:
            click.echo(click.style(f"\t{v}", fg="blue", bold=True))

        click.echo()

    ctx.exit(1 if stales else 0)


@graph.command()
@click.option("-n", "--dry-run", is_flag=True, help="Show steps that will be updated without running them.")
def update(dry_run):
    r"""Equivalent of `renku update`."""

    communicator = ClickCallback()
    perform_update().with_communicator(communicator).build().execute(dry_run=dry_run)


@graph.command()
@click.argument("path", type=click.Path(exists=False, dir_okay=False))
def save(path):
    r"""Save dependency graph as PNG."""
    with measure("CREATE DEPENDENCY GRAPH"):

        def _to_png(client, path):
            dg = DependencyGraph.from_json(client.dependency_graph_path)
            dg.to_png(path=path)

        Command().command(_to_png).build().execute(path=path)


@graph.command()
@click.option("--format", type=CaseInsensitiveChoice(FORMATS), default="json-ld", help="Choose an output format.")
@click.option("--strict", is_flag=True, default=False, help="Validate triples before output.")
@click.option("--workflows-only", is_flag=True, help="Exclude datasets metadata from export.")
def export(format, strict, workflows_only):
    r"""Equivalent of `renku log --format json-ld`."""
    communicator = ClickCallback()
    export_graph().with_communicator(communicator).build().execute(
        format=format, workflows_only=workflows_only, strict=strict
    )


@graph.group()
def dataset():
    """Proof-of-Concept command for dataset operations using new metadata."""


@dataset.command()
@click.argument("name")
@click.option("-t", "--title", default=None, type=click.STRING, help="Title of the dataset.")
@click.option("-d", "--description", default=None, type=click.STRING, help="Dataset's description.")
@click.option(
    "-c",
    "--creator",
    "creators",
    default=None,
    multiple=True,
    help="Creator's name, email, and affiliation. Accepted format is ``Forename Surname <email> [affiliation]``.",
)
@click.option("-k", "--keyword", default=None, multiple=True, type=click.STRING, help="List of keywords or tags.")
def create(name, title, description, creators, keyword):
    """Create a new dataset."""
    communicator = ClickCallback()

    result = (
        create_dataset()
        .with_communicator(communicator)
        .build()
        .execute(name=name, title=title, description=description, creators=creators, keywords=keyword)
    )

    new_dataset = result.output

    click.echo(f"Use the name ``{new_dataset.name}`` to refer to this dataset.")
    click.secho("OK", fg="green")


@dataset.command()
@click.argument("name")
@click.argument("urls", nargs=-1)
@click.option("-e", "--external", is_flag=True, help="Creates a link to external data.")
@click.option("--force", is_flag=True, help="Allow adding otherwise ignored files.")
@click.option("-o", "--overwrite", is_flag=True, help="Overwrite existing files.")
@click.option("-c", "--create", is_flag=True, help="Create dataset if it does not exist.")
@click.option("-s", "--source", "sources", default=None, multiple=True, help="Paths within remote git repo to be added")
@click.option("-d", "--destination", "destination", default="", help="Destination directory within the dataset path")
@click.option("--ref", default=None, help="Add files from a specific commit/tag/branch.")
def add(name, urls, external, force, overwrite, create, sources, destination, ref):
    """Add data to a dataset."""
    communicator = ClickCallback()

    add_to_dataset().with_communicator(communicator).build().execute(
        urls=urls,
        name=name,
        external=external,
        force=force,
        overwrite=overwrite,
        create=create,
        sources=sources,
        destination=destination,
        ref=ref,
    )
    click.secho("OK", fg="green")


@graph.group()
def workflow():
    """Proof-of-Concept command for workflow operations using new metadata."""


@workflow.command()
@click.argument("name", metavar="<name or uuid>")
@click.option("-f", "--force", is_flag=True, help="Force remove (don't prompt user to confirm).")
def remove(name, force):
    """Remove the workflow named <name>."""
    communicator = ClickCallback()
    remove_workflow().with_communicator(communicator).build().execute(name=name, force=force)



# -*- coding: utf-8 -*-
#
# Copyright 2018-2021 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test ``login`` command."""

import git
import pytest

from renku.cli import cli
from renku.core.commands.login import read_renku_token
from renku.core.utils.contexts import chdir
from tests.cli.fixtures.cli_gateway import ACCESS_TOKEN, ENDPOINT, USER_CODE


def test_login(runner, client_with_remote, mock_login):
    """Test login command."""
    remote_url = client_with_remote.repo.remotes[0].url

    result = runner.invoke(cli, ["login", "--git", "--yes", ENDPOINT], input=USER_CODE)

    assert 0 == result.exit_code
    assert ACCESS_TOKEN == read_renku_token(client_with_remote, ENDPOINT)
    credential = client_with_remote.repo.config_reader().get_value("credential", "helper")
    assert f"!renku token --hostname {ENDPOINT}" == credential
    assert {"origin", "renku-backup-origin"} == {r.name for r in client_with_remote.repo.remotes}
    assert remote_url == client_with_remote.repo.remotes["renku-backup-origin"].url
    assert client_with_remote.repo.remotes["origin"].url.startswith(f"https://{ENDPOINT}/repo")


@pytest.mark.parametrize("args", [[], ["--git"]])
def test_login_no_endpoint(runner, client, mock_login, args):
    """Test login command with no endpoint."""
    result = runner.invoke(cli, ["login"] + args)

    assert 2 == result.exit_code
    assert "Parameter 'endpoint' is missing." in result.output


@pytest.mark.parametrize("args", [[], ["--git"]])
def test_login_no_endpoint_and_remote(runner, client_with_remote, mock_login, args):
    """Test login command with no endpoint and with project remote."""
    result = runner.invoke(cli, ["login"] + args)

    assert 2 == result.exit_code, result.output
    assert "Parameter 'endpoint' is missing." in result.output


@pytest.mark.parametrize("args", [[], ["--git"]])
def test_login_invalid_endpoint(runner, client, mock_login, args):
    """Test login with and invalid endpoint."""
    result = runner.invoke(cli, ["login", "http: //example.com"] + args)

    assert 2 == result.exit_code
    assert "Invalid endpoint: `http: //example.com`." in result.output


def test_login_invalid_user_code(runner, client, mock_login):
    """Test login fails if user code is not valid."""
    result = runner.invoke(cli, ["login", ENDPOINT], input="invalid_user_code")

    assert 1 == result.exit_code
    assert "Remote host did not return an access token:" in result.output
    assert "status code: 404" in result.output


def test_login_with_config_endpoint(runner, client, mock_login):
    """Test login command with endpoint in config file."""
    assert 0 == runner.invoke(cli, ["config", "set", "endpoint", ENDPOINT]).exit_code

    result = runner.invoke(cli, ["login"], input=USER_CODE)

    assert 0 == result.exit_code
    assert "Successfully logged in." in result.output


def test_logout(runner, client, mock_login):
    """Test logout removes all credentials."""
    assert 0 == runner.invoke(cli, ["login", ENDPOINT], input=USER_CODE).exit_code

    result = runner.invoke(cli, ["logout"])

    assert 0 == result.exit_code
    assert read_renku_token(client, ENDPOINT) is None
    assert "Successfully logged out." in result.output


def test_repeated_login(runner, client, mock_login):
    """Test multiple logins."""
    assert 0 == runner.invoke(cli, ["login", ENDPOINT], input=USER_CODE).exit_code

    assert 0 == runner.invoke(cli, ["login", ENDPOINT], input=USER_CODE).exit_code
    assert ACCESS_TOKEN == read_renku_token(client, ENDPOINT)


def test_repeated_logout(runner, client, mock_login):
    """Test multiple logouts."""
    assert 0 == runner.invoke(cli, ["login", ENDPOINT], input=USER_CODE).exit_code

    assert 0 == runner.invoke(cli, ["logout"]).exit_code

    assert 0 == runner.invoke(cli, ["logout"]).exit_code
    assert read_renku_token(client, ENDPOINT) is None


def test_login_to_multiple_endpoints(runner, client, mock_login):
    """Test login to multiple endpoints."""
    second_endpoint, second_token = "second.endpoint", "second-token"
    mock_login.add_endpoint_token(second_endpoint, second_token)
    assert 0 == runner.invoke(cli, ["login", ENDPOINT], input=USER_CODE).exit_code

    assert 0 == runner.invoke(cli, ["login", second_endpoint], input=USER_CODE).exit_code

    assert ACCESS_TOKEN == read_renku_token(client, ENDPOINT)
    assert second_token == read_renku_token(client, second_endpoint)


def test_logout_all(runner, client, mock_login):
    """Test logout with no endpoint removes multiple credentials."""
    second_endpoint, second_token = "second.endpoint", "second-token"
    mock_login.add_endpoint_token(second_endpoint, second_token)
    assert 0 == runner.invoke(cli, ["login", ENDPOINT], input=USER_CODE).exit_code
    assert 0 == runner.invoke(cli, ["login", second_endpoint], input=USER_CODE).exit_code

    assert 0 == runner.invoke(cli, ["logout"]).exit_code

    assert read_renku_token(client, ENDPOINT) is None
    assert read_renku_token(client, second_endpoint) is None


def test_logout_one_endpoint(runner, client, mock_login):
    """Test logout from an endpoint removes credentials for that endpoint only."""
    second_endpoint, second_token = "second.endpoint", "second-token"
    mock_login.add_endpoint_token(second_endpoint, second_token)
    assert 0 == runner.invoke(cli, ["login", ENDPOINT], input=USER_CODE).exit_code
    assert 0 == runner.invoke(cli, ["login", second_endpoint], input=USER_CODE).exit_code

    assert 0 == runner.invoke(cli, ["logout", ENDPOINT]).exit_code

    assert read_renku_token(client, ENDPOINT) is None
    assert second_token == read_renku_token(client, second_endpoint)


def test_logout_non_existing_endpoint(runner, client, mock_login):
    """Test logout from a non-existing endpoint does nothing."""
    assert 0 == runner.invoke(cli, ["login", ENDPOINT], input=USER_CODE).exit_code

    assert 0 == runner.invoke(cli, ["logout", "non.existing"]).exit_code

    assert read_renku_token(client, ENDPOINT) is not None


def test_login_git_abort(runner, client_with_remote):
    """Test login command."""
    result = runner.invoke(cli, ["login", "--git", ENDPOINT], input="n")

    assert 1 == result.exit_code
    assert "Remote URL will be changed. Do you want to continue?" in result.output
    assert "Aborted!" in result.output


def test_login_non_git(runner, client, directory_tree):
    """Test login from a non-git directory."""
    with chdir(directory_tree):
        result = runner.invoke(cli, ["login", "--git", ENDPOINT])

    assert 2 == result.exit_code
    assert "Cannot use '--git' flag outside a project" in result.output


def test_login_git_no_unique_remote(runner, client_with_remote):
    """Test login from a git directory with no clear remote."""
    client_with_remote.repo.create_remote("second-remote", "second-remote.net")
    client_with_remote.repo.create_head("branch-with-no-remote")
    client_with_remote.repo.heads["branch-with-no-remote"].checkout()

    result = runner.invoke(cli, ["login", "--git", ENDPOINT])

    assert 2 == result.exit_code
    assert "Cannot find a unique remote URL for project." in result.output


def test_repeated_git_login(runner, client_with_remote, mock_login):
    """Test multiple logins to git repo fails to change remote URL after first time."""
    remote_url = client_with_remote.repo.remotes[0].url

    assert 0 == runner.invoke(cli, ["login", "--git", "--yes", ENDPOINT], input=USER_CODE).exit_code

    result = runner.invoke(cli, ["login", "--git", "--yes", ENDPOINT], input=USER_CODE)

    assert 0 == result.exit_code
    assert "Backup remove 'renku-backup-origin' already exists. Ignoring '--git' flag." in result.output
    assert "Error: Cannot create backup remote 'renku-backup-origin' for" not in result.output
    assert {"origin", "renku-backup-origin"} == {r.name for r in client_with_remote.repo.remotes}
    assert remote_url == client_with_remote.repo.remotes["renku-backup-origin"].url
    assert client_with_remote.repo.remotes["origin"].url.startswith(f"https://{ENDPOINT}/repo")
    assert not client_with_remote.repo.remotes["origin"].url.startswith(f"https://{ENDPOINT}/repo/repo")


def test_logout_git(runner, client_with_remote, mock_login):
    """Test logout removes backup remotes and restores original remote url."""
    remote_url = client_with_remote.repo.remotes[0].url

    assert 0 == runner.invoke(cli, ["login", "--git", "--yes", ENDPOINT], input=USER_CODE).exit_code

    result = runner.invoke(cli, ["logout"])

    assert 0 == result.exit_code
    assert {"origin"} == {r.name for r in client_with_remote.repo.remotes}
    assert remote_url == client_with_remote.repo.remotes["origin"].url
    try:
        credential = client_with_remote.repo.git.config("credential.helper", local=True)
    except git.exc.GitCommandError:  # NOTE: If already logged out, ``git config --unset`` raises an exception
        credential = None
    assert credential is None


def test_token(runner, client_with_remote, mock_login):
    """Test get credential when valid credential exist."""
    assert 0 == runner.invoke(cli, ["login", ENDPOINT], input=USER_CODE).exit_code

    result = runner.invoke(cli, ["token", "--hostname", ENDPOINT, "get"])

    assert 0 == result.exit_code
    assert "username=renku\n" in result.output
    assert f"password={ACCESS_TOKEN}\n" in result.output


def test_token_non_existing_hostname(runner, client_with_remote, mock_login):
    """Test get credential for a different hostname."""
    assert 0 == runner.invoke(cli, ["login", ENDPOINT], input=USER_CODE).exit_code

    result = runner.invoke(cli, ["token", "--hostname", "non-existing", "get"])

    assert 0 == result.exit_code
    assert "username=renku\n" in result.output
    assert "password=\n" in result.output


def test_token_no_credential(runner, client_with_remote, mock_login):
    """Test get credential when valid credential doesn't exist."""
    assert 0 == runner.invoke(cli, ["logout"]).exit_code

    result = runner.invoke(cli, ["token", "--hostname", ENDPOINT, "get"])

    assert 0 == result.exit_code
    assert "username=renku\n" in result.output
    assert "password=\n" in result.output


def test_token_invalid_command(runner, client_with_remote, mock_login):
    """Test call credential helper with a command other than 'get'."""
    assert 0 == runner.invoke(cli, ["login", ENDPOINT], input=USER_CODE).exit_code

    result = runner.invoke(cli, ["token", "--hostname", ENDPOINT, "non-get-command"])

    assert 0 == result.exit_code
    assert "" == result.output
